softplus <- function(x){
  y <- rep(NaN,length(x))
  if(any(x>0)){
    y[x>0] <- log(1+exp(-x[x>0]))
    if (any(is.finite(y)))
      y[is.finite(y)] <- y[is.finite(y)]+x[is.finite(y)]
     }
    if(any(x<=0)){
      y[x<=0]=log(1+exp(x[x<=0]))
    if (any(!is.finite(y)))
      y[!is.finite(y)] <- 0
    }
    y
}

softplusinv <- function(y){
  log(exp(y)-1)
}

lambertw <- function(z){
  w0 <- .C("lambertwR", as.double(z), results=double(1),PACKAGE="hparetomixt")
  w0[["results"]]
}

hpareto.nll <- function(theta,x){
  # assume that the 3rd parameter is encoded as log(sigma)
  if (length(theta)!=3)
    stop("THETA must have length 3")
  n <- length(x)
  if (n == 0)
    stop("X must have a positive length")
  results <- .C("hpnll",as.double(theta),as.double(x),as.integer(n),nll=double(1),nllgrad=double(3),PACKAGE="hparetomixt")

  obj <- results[["nll"]]
  attr(obj,"gradient") <-  results[["nllgrad"]]
  obj
}


hpareto.fit <- function(params,x,...){
  theta <- params
  theta[3] <- log(params[3])
  opt <- nlm(hpareto.nll,theta,x,...)
  params.opt <- opt$estimate
  params.opt[3] <- exp(params.opt[3])
  params.opt
}


hpareto.negloglike <- function(params,x){
  # params is a vector of the three hybrid Pareto parameters xi, mu and sigma
  theta <- params
  theta[3] <- log(params[3])
  hpareto.nll(theta,x)
}

hillest <- function(data,k){
  # Computes the Hill estimator of the tail index xi
  # The threshold used is the K+1 order statistic

  xstat <- sort(data, decreasing = TRUE)
  xihat <- mean(log(xstat[1:k])) - log(xstat[k+1])
  xihat
}


gpd.mme <- function(x){
  #  Moments estimator for the GPD
  xbar <- mean(x)
  s2 <- var(x)
  s2 <- s2 + (s2==0)
  xi0 <- -0.5 * (((xbar * xbar)/s2) - 1)
  beta0 <- 0.5 * xbar * (((xbar * xbar)/s2) + 1)
  c(xi0,beta0)
}


hpareto.mme <- function(x,xi0=c(),p=0.99){
  # equivalent to sgpdmmepos.m, initialises the parameters of a hybrid Pareto mixture based on the method of moments and the data in X, no truncation here
  # more doc to be found in that file /Users/julie/diro/evt/Mfiles/mystats/sgpdmmepos.m

  k <- length(x) - round(p*length(x)) + 1  # use the 10% largest data to estimate xi

  npos <- sum(x > 0) # ensure the k+1 largest values are positive
  if (npos < k + 1)
    k <- npos -1

  if (length(xi0)==0){ # xi0 is not given externally
    xi0 <- hillest(x, k)
    if (xi0 < 0) # enforce positivity of the tail index
      xi0 <- 0
  }

  z <- (1 + xi0)^2/(2 * pi)
  w <- lambertw(z)
  phiw <- pnorm(sign(1 + xi0) * sqrt(w))

  mu0 <- quantile(x, 1/(2 * (1 + phiw)))
  alpha0 <- quantile(x, phiw/(1 + phiw))
  names(mu0) <- NULL
  names(alpha0) <- NULL
  sigma0 <- (alpha0 - mu0)/ (sqrt(w) * sign(1 + xi0))

  if (sigma0 <= 0)
    if (sum(x > alpha0) == 0)
      sigma0 <- 1
    else{
      theta <- gpd.mme(x[x > alpha0] - alpha0)
      sigma0 <- sqrt(w) * theta[2]/sqrt((1+xi0)^2)
    }
  c(xi0,mu0,sigma0)
}


hpareto.beta <- function(xi,sigma=1){
  sigma * sqrt((1 + xi)^2) / sqrt(lambertw((1 + xi)^2 / (2 * pi)))
}


hpareto.alpha <- function(xi,mu=0,sigma=1){
  mu + sign(1 + xi) * sigma * sqrt(lambertw((1 + xi)^2 / (2 * pi)))
}

hpareto.gamma <- function(xi,mu=0,sigma=1,trunc=T){
  alpha <- hpareto.alpha(xi,mu,sigma)
  if (trunc){
    if (alpha > 0)
      return(1 + pnorm(sign(1 + xi) * sqrt(lambertw((1 + xi)^2 / (2 * pi))))-pnorm(-mu/sigma))
    else{
      beta <- hpareto.beta(xi,sigma)
      return(1-pgpd(-alpha,xi,beta))
    }
  } else
    return(1 + pnorm(sign(1 + xi) * sqrt(lambertw((1 + xi)^2 / (2 * pi)))))
}

dhpareto <- function(y,xi,mu=0,sigma=1,log=FALSE,trunc=TRUE){
  # hybrid Pareto density function
  w <- lambertw((1 + xi)^2 / (2 * pi))
  beta <- sigma * sqrt((1 + xi)^2) / sqrt(w)
  alpha <- sign(1 + xi) * sigma * sqrt(w) + mu

  p <- rep(0,length(y))
  if (trunc){
    if (alpha > 0){  # truncation includes the Gaussian part
      gamma <-  1 + pnorm(sign(1 + xi) * sqrt(lambertw((1 + xi)^2 / (2 * pi))))-pnorm(-mu/sigma)
      if (any(y>0 & y<=alpha))
        p[which(y>0 & y<=alpha)] <- dnorm(y[which(y>0 & y<=alpha)],mu,sigma,log)
    }
    else
      gamma <- 1-pgpd(0,alpha,beta,xi) # no points in the Gaussian part
    if (any(y>alpha & y > 0))
      p[which(y>alpha & y > 0)] <- dgpd(y[which(y>alpha & y > 0)],alpha,beta,xi,log)
  } else { # no truncation
    gamma <-  1 + pnorm(sign(1 + xi) * sqrt(lambertw((1 + xi)^2 / (2 * pi))))
    if (any(y<=alpha))
      p[which(y<=alpha)] <- dnorm(y[which(y<=alpha)],mu,sigma,log)
     if (any(y>alpha))
       p[which(y>alpha)] <- dgpd(y[which(y>alpha)],alpha,beta,xi,log)
  }
  if (log)
    p-log(gamma)
  else
    p/gamma
}


phpareto <- function(q,xi,mu=0,sigma=1,trunc=TRUE){
  # hybrid Pareto distribution function
  w <- lambertw((1 + xi)^2 / (2 * pi))
  beta <- sigma * sqrt((1 + xi)^2) / sqrt(w)
  alpha <- sign(1 + xi) * sigma * sqrt(w) + mu

  p <- rep(0,length(q))
  if (trunc){
    if (alpha > 0){ # truncation includes the Gaussian part
      gamma <-  1 + pnorm(sign(1 + xi) * sqrt(lambertw((1 + xi)^2 / (2 * pi))))-pnorm(-mu/sigma)
      if (any(q>0 & q<=alpha))
        p[which(q>0 & q<=alpha)] <- pnorm(q[which(q>0 & q<=alpha)],mu,sigma)-pnorm(0,mu,sigma)
      if (any(q>alpha))
        p[which(q>alpha)] <- pgpd(q[which(q>alpha)],alpha,beta,xi)+pnorm(alpha,mu,sigma)-pnorm(0,mu,sigma)
    } else{ # if alpha < 0 and trunc = T, then all the points are in the GPD part
      gamma <- 1-pgpd(0,alpha,beta,xi)
      if (any(q>alpha & q > 0))
        p[which(q>alpha & q > 0)] <- pgpd(q[which(q>alpha & q > 0)],alpha,beta,xi)-pgpd(0,alpha,beta,xi)
    }
  } else{ # no truncation
    gamma <-  1 + pnorm(sign(1 + xi) * sqrt(lambertw((1 + xi)^2 / (2 * pi))))
    if (any(q<=alpha))
      p[which(q<=alpha)] <- pnorm(q[which(q<=alpha)],mu,sigma)
    if (any(q>alpha))
      p[which(q>alpha)] <- pgpd(q[which(q>alpha)],alpha,beta,xi)+pnorm(alpha,mu,sigma)
  }

  p/gamma
}


rhpareto <- function(n,xi,mu=0,sigma=1,trunc=TRUE){
  w <- lambertw((1 + xi)^2 / (2 * pi))
  beta <- sigma * sqrt((1 + xi)^2) / sqrt(w)
  alpha <- sign(1 + xi) * sigma * sqrt(w) + mu

  u <- runif(n)
  r <- rep(NaN,n)

  if (trunc){
    if (alpha > 0){  # truncation includes the Gaussian part
      F0 <- pnorm(0,mu,sigma) # probability of being below zero for a Gaussian variable
      Falpha <- pnorm(alpha,mu,sigma) # probability of being below alpha for a Gaussian
      Halpha <- (Falpha-F0)/(1+Falpha-F0) # probability of being in the positive Gaussian part
      if (any(u<=Halpha)) # generate data in the positive Gaussian part
        r[which(u<=Halpha)] <- qnorm(runif(sum(u<=Halpha))*(Falpha-F0)+F0,mu,sigma)
      if (any(u>Halpha)) # generate data in the GPD part
        r[which(u>Halpha)] <- qgpd(runif(sum(u>Halpha)),alpha,beta,xi)
    } else { # no points in the Gaussian part
      Galpha <- pgpd(0,alpha,beta,xi) # prob of being in the negative part of the GPD
      r <- qgpd(u*(1-Galpha)+Galpha,alpha,beta,xi)
    }
  }
  else{  # no truncation
    Falpha <- pnorm(sign(1 + xi) * sqrt(lambertw((1 + xi)^2 / (2 * pi))))
    gamma <-  1 + Falpha
    Halpha <- Falpha/gamma   # probability of being in the Gaussian part
    if (any(u<=Halpha)) # generate data in the Gaussian part
      r[which(u<=Halpha)] <- qnorm(runif(sum(u<=Halpha))*Falpha,mu,sigma)
    if (any(u>Halpha)) # generate data in the GPD part
      r[which(u>Halpha)] <- qgpd(runif(sum(u>Halpha)),alpha,beta,xi)
  }

  r
}


qhpareto <- function(p,xi,mu=0,sigma=1,trunc=TRUE){
  w <- lambertw((1 + xi)^2 / (2 * pi))
  beta <- sigma * sqrt((1 + xi)^2) / sqrt(w)
  alpha <- sign(1 + xi) * sigma * sqrt(w) + mu
  Falpha <- pnorm((alpha-mu)/sigma)

  y <- rep(0,length(p))
  if (trunc)
    F0 <- pnorm(0,mu,sigma)
  else
    F0 <- 0

  Halpha <-  (Falpha-F0)/(1+Falpha-F0)
  if (any(p<=Halpha))
    y[p<=Halpha] <- qnorm((1+Falpha-F0)*p[p<=Halpha]+F0,mu,sigma)

  if (any(p>Halpha))
    y[p>Halpha] <- qgpd((1+Falpha-F0)*p[p>Halpha]-Falpha+F0,0,beta,xi)+alpha
  y
}


hparetomixt.init <- function(m,x,iter.max=20,nstart=10){
  # initialize a mixture of hybrid Paretos
  # equivalent to sgpdmminit.m but do not use a structure MIX, parameters of
  # the model are stored in params, 4 x m matrix

  ndata <- length(x) # assumes the data is univariate
  params.init <- matrix(nrow=4,ncol=m)

  if (m>1){
    clustering <- kmeans(x,m,iter.max,nstart)
    params.init[3,] <- clustering$centers

    #  Set priors depending on number of points in each cluster
    cluster.sizes <- sapply(clustering$size,max,1) # Make sure that no prior is zero
    params.init[1,] <- cluster.sizes/sum(cluster.sizes) # Normalise priors

     # Hybrid Pareto parameters are estimated on each cluster making sure that the xi (tail index parameters) are positive
    minel <- 10 # minimum number of points to perform density estimation on a component
    for (j in 1:m){
      if (clustering$size[j] < minel){ # not enough points in the cluster, initialise parameters randomly
        params.init[2,j] <- runif(1)*0.5 # initialise a small tail index (=thin tail)
        params.init[4,j] <- runif(1)*10+10
      } else {
        theta <- hpareto.mme(x[which(clustering$cluster==j)])
        params.init[2,j] <- theta[1]
        params.init[3,j] <- theta[2]
        params.init[4,j] <- theta[3]
      }

    }
  } else { # there is just one cluster
    theta <- hpareto.mme(x)
    params.init[1,1] <- 1
    params.init[2,1] <- theta[1]
    params.init[3,1] <- theta[2]
    params.init[4,1] <- theta[3]
  }
  if(any(params.init[4,]<0.001))
    params.init[4,params.init[4,]<0.001] <- 0.001
  if(any(params.init[2,]<0.001))
    params.init[2,params.init[2,]<0.001] <- 0.001
  params.init
}






dhparetomixt <- function(params,x,log=FALSE,trunc=TRUE){
  # params is 4 x m matrix
  m <- dim(params)[2]
  p <- rep(0,length(x))
  for (j in 1:m){
    p <- p+params[1,j]*dhpareto(x,params[2,j],params[3,j],params[4,j],log=F,trunc=FALSE)
  }
  if (trunc){
    F0 <- phparetomixt(params,0,trunc=FALSE)
    p <- p/(1-F0)
    if (any(x<=0))
      p[which(x<=0)] <- 0
  }

  if (log)
    p <- log(p)
  p
}

phparetomixt <- function(params,x,trunc=TRUE){
  # params is 4 x m matrix
  m <- dim(params)[2]
  p <- rep(0,length(x))
  for (j in 1:m){
    p <- p+params[1,j]*phpareto(x,params[2,j],params[3,j],params[4,j],trunc=FALSE)
  }
  if (trunc){
    F0 <- phparetomixt(params,0,trunc=FALSE)
    p <- p/(1-F0)
    if (any(x<=0))
      p[which(x<=0)] <- 0
  }
  p
}


hparetomixt.nll <- function(theta,m,x){
  results <- .C("ummhnll",as.double(theta),as.integer(m),as.double(x),as.integer(length(x)),nll=double(1),nllgrad=double(length(theta)),PACKAGE="hparetomixt")

  obj <- results[["nll"]]
  attr(obj,"gradient") <-  results[["nllgrad"]]
  obj
}

hparetomixt.negloglike <- function(params,x){
  # params is an 4 by m matrix
  m <- dim(params)[2]
  theta <- ummhbwd(params,m)
  hparetomixt.nll(theta,m,x)
}

ummhbwd <- function(params,m){
  cfunction <- .C("ummhbwd",as.double(t(params)), as.integer(m), theta=double(4*m-1),PACKAGE="hparetomixt")
  cfunction[["theta"]]
}

hparetomixt.fwd <- function(theta,m){
  cfunction <- .C("ummhfwd",as.double(theta), as.integer(m), params=double(4*m),PACKAGE="hparetomixt")
  matrix(cfunction[["params"]],4,m,byrow=TRUE)
}

hparetomixt.fit <- function(params,x,...){
  # params0 is an 4 by m matrix
  m <- dim(params)[2]
  theta0 <- ummhbwd(params,m)
  opt <- nlm(hparetomixt.nll,theta0,m,x,...)
  hparetomixt.fwd(opt$est,m)
}


hparetomixt.cvtrain <- function(m,x,nfold=5,nstart=1,...){
  n <- length(x)
  x <- x[sample(1:n,n)]  # shuffle data


  nk <- floor(n/nfold)  # number of observations per fold
  nlltest <- 0
  for (k in 1:nfold){ # train and test on each fold
    if (k < nfold)
      indk <- (1+(k-1)*nk):(k*nk)
    else
      indk <-  (1+(k-1)*nk):max(k*nk,n)
    cat("Training fold ", k ,"\t")
    nllbest <- Inf
    for (i in 1:nstart){
      params.init <- hparetomixt.init(m,x[-indk])
      theta <- ummhbwd(params.init,m)
      opt <- nlm(hparetomixt.nll,theta,m,x[-indk],...)
      if (opt$min<nllbest){
        nllbest <- opt$min
        thetabest <- opt$est
      } # if
    } # for nstart
    nlltest <- nlltest+hparetomixt.nll(thetabest,m,x[indk])
    cat("Test error is :",nlltest,"\n")
  }  # for nfold
  attr(nlltest,"gradient") <- NULL
  nlltest
}


hparetomixt.negloglike.tailpen <- function(params,lambda,w,beta,mu,sigma,x){
  m <- dim(params)[2]
  theta <- ummhbwd(params,m)
  hparetomixt.nll.bimodal.tailpen(theta,m,lambda,w,beta,mu,sigma,x)
}

hparetomixt.nll.bimodal.tailpen <- function(theta,m,lambda,w,beta,mu,sigma,x){
  results <- .C("ummhnll_bimodal_tailpen",as.double(theta),as.integer(m),as.double(lambda),
                as.double(w),as.double(beta),as.double(mu),as.double(sigma),as.double(x),
                as.integer(length(x)),nll=double(1),nllgrad=double(length(theta)),PACKAGE="hparetomixt")
  obj <- results[["nll"]]
  attr(obj,"gradient") <-  results[["nllgrad"]]
  obj
}

hparetomixt.fit.tailpen <- function(params,lambda,w,beta,mu,sigma,x,...){
  # params0 is an 4 by m matrix
  m <- dim(params)[2]
  theta0 <- ummhbwd(params,m)
  opt <- nlm(hparetomixt.nll.bimodal.tailpen,theta0,m,lambda,w,beta,mu,sigma,x,...)
  hparetomixt.fwd(opt$estimate,m)
}


hparetomixt.cvtrain.tailpen <- function(m,lambda,w,beta,mu,sigma,x,nfold=5,nstart=1,...){
  n <- length(x)
  nk <- floor(n/nfold)  # number of observations per fold
  x <- x[sample(1:n,n)]  # shuffle data

  nlltest <- 0
  for (k in 1:nfold){ # train and test on each fold
    if (k < nfold)
      indk <- (1+(k-1)*nk):(k*nk)
    else
      indk <-  (1+(k-1)*nk):max(k*nk,n)
    cat("Training fold ", k ,"\t")
    nllbest <- Inf
    for (i in 1:nstart){
      params.init <- hparetomixt.init(m,x[-indk])
      theta <- ummhbwd(params.init,m)
      opt <- nlm(hparetomixt.nll.bimodal.tailpen,theta,m,lambda,w,beta,mu,sigma,x[-indk],...)
      if (opt$min<nllbest){
        nllbest <- opt$min
        thetabest <- opt$est
      } # if
    } # for nstart
    nlltest <- nlltest+hparetomixt.nll(thetabest,m,x[indk])

    cat("Test error is :",nlltest,"\n")

  }
  attr(nlltest,"gradient") <- NULL
  nlltest
}

hparetomixt.disp <- function(params){
  # params0 is an 4 by m matrix
  m <- dim(params)[2]
  for (j in 1:m){
    cat(paste(j, ": &", format(params[1,j],digits=4), " &", format(params[2,j],digits=4), " &", format(params[3,j],digits=4), " &", format(params[4,j],digits=4), "\\\\", "\n"))
  }
}


gaussmixt.init <- function(m,x,iter.max=20,nstart=10){
  # initialize a mixture of Gaussians
  # the model are stored in params, m x 3 matrix

  ndata <- length(x) # assumes the data is univariate
  params.init <- matrix(nrow=3,ncol=m)

  clustering <- kmeans(x,m,iter.max,nstart)
  params.init[2,] <- clustering$centers

  #  Set priors depending on number of points in each cluster
  cluster.sizes <- sapply(clustering$size,max,1) # Make sure that no prior is zero
  params.init[1,] <- cluster.sizes/sum(cluster.sizes) # Normalise priors

# Gaussian standard deviations are estimated on each cluster
  for (j in 1:m){
    params.init[3,j] <- mad(x[which(clustering$cluster==j)])
  }
  if(any(params.init[3,]<0.1))
    params.init[3,which(params.init[3,]<0.1)] <- 0.101
  params.init
}


ummgbwd <- function(params,m){
  cfunction <- .C("ummgbwd",as.double(t(params)), as.integer(m), theta=double(3*m-1),PACKAGE="hparetomixt")
  cfunction[["theta"]]
}


pgaussmixt <- function(params,x,trunc=TRUE){
  # params is 3 x m matrix
  m <- dim(params)[2]
  p <- rep(0,length(x))
  for (j in 1:m){
    p <- p+params[1,j]*pnorm(x,params[2,j],params[3,j])
  }
  if (trunc){
    F0 <- pgaussmixt(params,0,trunc=FALSE)
    p <- p/(1-F0)
    if (any(x<=0))
      p[which(x<=0)] <- 0
  }
  p
}

dgaussmixt <- function(params,x,log=FALSE,trunc=TRUE){
  # params is 3 x m matrix
  m <- dim(params)[2]
  p <- rep(0,length(x))
  for (j in 1:m){
    p <- p+params[1,j]*dnorm(x,params[2,j],params[3,j])
  }

  if (trunc){
    F0 <- pgaussmixt(params,0,trunc=FALSE)
    p <- p/(1-F0)
    if (any(x<=0))
      p[which(x<=0)] <- 0
  }
  if (log)
    p <- log(p)
  p
}

dlognormixt <- function(params,x,log=FALSE){
  # params is 3 x m matrix
  m <- dim(params)[2]
  p <- rep(0,length(x))
  for (j in 1:m){
    p <- p+params[1,j]*dlnorm(x,params[2,j],params[3,j])
  }
  if (log)
    p <- log(p)
  p
}


