#include "hparetomixt.h"
     
/* Lambert W function. */
double lambertw(double z){
  double tolrel, tolabs, w0, errabs, errrel, e0, wj;
  tolrel = pow(10.0,-6.0);
  tolabs = tolrel;
  w0 = 1.0/2.0;
  errabs = fabs(z - w0 * exp(w0));
  errrel = errabs/z;

  while ((errabs > tolabs) & (errrel > tolrel)){
    e0 = exp(w0);
    wj = w0-(w0*e0-z)/(e0*(w0+1)-(w0+2)*(w0*e0-z)/(2*w0+2));
    w0 = wj;
    /*    mexPrintf("w0 = %f\n",w0);*/
    errabs = fabs(z-w0*exp(w0));
    errrel = errabs/z;
  }
  return(w0);
}

/* Lambert W function for use with R */
void lambertwR(double *z, double *w0){
  double tolrel, tolabs, errabs, errrel, e0, wj;
  tolrel = pow(10.0,-6.0);
  tolabs = tolrel;
  *w0 = 1.0/2.0;
  errabs = fabs(*z - *w0 * exp(*w0));
  errrel = errabs/ *z;

  while ((errabs > tolabs) & (errrel > tolrel)){
    e0 = exp(*w0);
    wj = *w0-(*w0*e0-*z)/(e0*(*w0+1)-(*w0+2)*(*w0*e0-*z)/(2*(*w0)+2));
    *w0 = wj;
    /*    mexPrintf("w0 = %f\n",*w0);*/
    errabs = fabs(*z-*w0*exp(*w0));
    errrel = errabs/ *z;
  }

}

/* Complementary error function */
double erfc(double x){
  double t,z,ans;
  z=fabs(x); 
  t=1.0/(1.0+0.5*z); 
  ans=t*exp(-z*z-1.26551223+t*(1.00002368+t*(0.37409196+t*(0.09678418+ 
       t*(-0.18628806+t*(0.27886807+t*(-1.13520398+t*(1.48851587+
      	t*(-0.82215223+t*0.17087277))))))))); 
	return x>= 0.0 ? ans : 2.0-ans; 		  
}

/* Error function */
double erf(double x){
  double t,z,ans;
  z=fabs(x); 
  t=1.0/(1.0+0.5*z); 
  ans=t*exp(-z*z-1.26551223+t*(1.00002368+t*(0.37409196+t*(0.09678418+ 
       t*(-0.18628806+t*(0.27886807+t*(-1.13520398+t*(1.48851587+
      	t*(-0.82215223+t*0.17087277))))))))); 
  return x>= 0.0 ? 1.0-ans : ans-1.0; 		  
}

/* Sign function */
double sign(double x){
  double y;
  if (x > 0.0)
    y = 1.0;
  else{
    if (x < 0.0)
      y = -1.0;
    else
      y = 0.0;		
  }
  return(y);
}

void hpnll(double *theta, double *x, int *n, double *nll, double *nllgrad){
  int i;
  double w, gamma, beta, alpha, xi, mu, sigma;
  double dgdx, dbdx, dbds, dadx, dads;

  xi=theta[0];
  mu=theta[1];
  sigma=exp(theta[2]);

  w = lambertw((1.0 + xi)*(1.0 + xi)/(2.0*PI));
  gamma = 3.0/2.0 + erf(sign(1.0+xi)*sqrt(w/2.0))/2.0;
  alpha = sign(1.0+xi) * sigma * sqrt(w) + mu;

  /* Make sure initialization of neg-log-like and its gradient
     is zero */
  *nll=0.0;
  nllgrad[0]=0;
  nllgrad[1]=0;
  nllgrad[2]=0;
  
    
  /* Compute partial derivatives that do not depend on data points:
     derivatives of gamma, beta and alpha with respect to xi, mu, sigma */
  dgdx=(exp(-w/2.0) * sqrt(w/(2.0*PI))) / (fabs(1.0 + xi) * (1.0 + w));
  dbdx=sign(1.0 + xi) * sigma * sqrt(w) / (1.0 + w);
  dbds=fabs(1.0 + xi) / sqrt(w);
  dadx=sigma * sqrt(w) / (fabs(1.0 + xi) * (1.0 + w));
  dads=sign(1.0 + xi) * sqrt(w);
 

  for (i=0; i<*n; i++){

    /* Compute neg-log-like  and its gradient */
    if (x[i] <= alpha){
	*nll+=log(gamma*sigma)+(x[i]-mu)*(x[i]-mu)/(2.0*sigma*sigma)+log(2*PI)/2;
	nllgrad[0]+=dgdx/gamma;
	nllgrad[1]+=-(x[i]-mu)/(sigma*sigma);
	nllgrad[2]+=1.0-(x[i]-mu)*(x[i]-mu)/(sigma*sigma);
    } 
    else {
      beta = sigma*fabs(1.0+xi)/sqrt(w);
      if (xi == 0.0){
	  *nll+= log(gamma*beta)+(x[i]-alpha)/beta;
	  nllgrad[0]+=dgdx/gamma + (dbdx*(1.0-(x[i]-alpha)/beta)-dadx)/beta;
	  nllgrad[1]+=-1.0/beta;
	  nllgrad[2]+=sigma/beta *((1.0-(x[i]-alpha)/beta)*dbds-dads);
      }
      else if (xi > 0.0 || (xi < 0.0 && (x[i] < alpha - beta/xi))){
	  *nll+=log(gamma*beta)+(1.0/xi + 1.0)*log(1.0+xi*(x[i]-alpha)/beta);
	  nllgrad[0]+=dgdx/gamma + dbdx/beta * 
	    (1.0-(xi+1.0)*(x[i]-alpha)/(beta+xi*(x[i]-alpha))) - 
	    log(1.0+xi*(x[i]-alpha)/beta)/(xi*xi) + 
	    (1.0+1.0/xi)*(x[i]-alpha)/(beta+xi*(x[i]-alpha)) -
	    (1.0+xi)/(beta+xi*(x[i]-alpha))*dadx;
	  nllgrad[1]+=-(1.0+xi)/(beta+xi*(x[i]-alpha));
	  nllgrad[2]+=(dbds/beta * 
		       (1.0-(xi+1.0)*(x[i]-alpha)/(beta+xi*(x[i]-alpha))) - 
		       (xi+1.0)/(beta+xi*(x[i]-alpha)) * dads)*sigma;
      }
      else{
	  *nll+=INFINITY;
	  nllgrad[0]+=INFINITY;
	  nllgrad[1]+=INFINITY;
	  nllgrad[2]+=INFINITY;
      }
    }

  } /* for i */
}



/* Univariate Normal density N(mu,sigma^2) evaluated at x. */
double normpdf(double mu, double sigma, double x){
  double z = (x-mu)/sigma;
  return(exp(-z*z/2.0)/(sqrt(2.0*PI)*sigma));
}


/* Univariate Normal distribution fonction N(mu,sigma^2) evaluated at x. */
double normcdf(double mu, double sigma, double x){
  double erfc(double x);
  double z = (x-mu)/sigma;
  return(0.5 * erfc(-z/sqrt(2)));
}


/* Logarithm of the univariate Normal density N(mu,sigma^2) evaluated at x. */
double normlogpdf(double mu, double sigma, double x){
  double z = (x-mu)/sigma;
  return(-z*z/2.0 - log(2.0*PI)/2.0 -log(sigma));
}



/* Generalized Pareto density G(x;xi,beta) evaluated at x. */
double gpdpdf(double xi, double beta, double x){
  double z;
  /* Check domain conditions. */
  if ((x < 0.0) | ((xi < 0.0) & (x > -beta/xi)) )
    return(0.0);
  z = x/beta;
  if (xi == 0.0)
    return(exp(-z)/beta);
  else{
    return(pow((1.0 + xi * z),(-1.0/xi - 1.0))/beta);
  }
}


/* Logarithm of the Generalized Pareto density G(xi,beta) evaluated at x. */
double gpdlogpdf(double xi, double beta, double x){
  double z;
  /* Check domain conditions. */
  if ((x < 0.0) | ((xi < 0.0) & (x > -beta/xi)) )
    return(INFINITY);
  z = x/beta;
  if (xi == 0.0)
    return(-z-log(beta));
  else{
    /*if (xi > 0.0)*/
    return((-1.0/xi - 1.0)*log(1.0 + xi * z)-log(beta));
  }
}

/* Hybrid Pareto density h(x; xi, mu, sigma) evaluated at one point. */
double hpdf(double xi, double mu, double sigma, double x){
  double w, gamma, beta, alpha,y;
  
  w = lambertw((1.0 + xi)*(1.0 + xi)/(2.0*PI));
  alpha = sign(1.0+xi) * sigma * sqrt(w) + mu;
  gamma = 3.0/2.0 + erf(sign(1.0+xi)*sqrt(w/2.0))/2.0;

  if (x <= alpha){
    y=normpdf(mu, sigma, x)/gamma;
  }
  else{
    beta = sigma*fabs(1.0+xi)/sqrt(w);
    y=gpdpdf(xi, beta, x-alpha)/gamma;    
  }
  
  return(y);
}


/* Hybrid Pareto cdf h(x; xi, mu, sigma) evaluated at one point. */
double hcdf(double xi, double mu, double sigma, double x){
  double w, gamma, beta, alpha,y;
  
  w = lambertw((1.0 + xi)*(1.0 + xi)/(2.0*PI));
  gamma = 3.0/2.0 + erf(sign(1.0+xi)*sqrt(w/2.0))/2.0;
  alpha = sign(1.0+xi) * sigma * sqrt(w) + mu;
  
  if (x <= alpha){
    y=normcdf(mu, sigma, x)/gamma;
  }
  else{
    beta = sigma*fabs(1.0+xi)/sqrt(w);
    if (xi == 0.0)
      y=(normcdf(mu,sigma,alpha)+1.0-exp(-(x-alpha)/beta))/gamma;
    else 
      y=(normcdf(mu,sigma,alpha)+1.0-pow(1.0+xi*(x-alpha)/beta,-1/xi))/gamma;
  }
  
  return(y);
}

/* Logarithm of the Hybrid Pareto density h(x; xi, mu, sigma) 
   evaluated at one point. */ 
double hlogpdf(double xi, double mu, double sigma, double x){
  double w, gamma, beta, alpha,y;
  
  w = lambertw((1.0 + xi)*(1.0 + xi)/(2.0*PI));
  gamma = 3.0/2.0 + erf(sign(1.0+xi)*sqrt(w/2.0))/2.0;
  alpha = sign(1.0+xi) * sigma * sqrt(w) + mu;
  
  if (x <= alpha){
    y=normlogpdf(mu, sigma, x)-log(gamma);
  }
  else{
    beta = sigma*fabs(1.0+xi)/sqrt(w);
    y=gpdlogpdf(xi, beta, x-alpha)-log(gamma);    
  }
  return(y);
}



/* --------------------- UMMH : Unconditional Mixture Model -------------------
   -----------------------  With Hybrid Pareto components   -------------------*/

/* Softplus function ensures positivity: y = log(1+exp(x)) */
double softplus(double x){
  double y;
  if(x>0.0){
    y = log(1.0+exp(-x));
    if (isfinite(y) & !isnan(y))
      y+=x;
    else
      y=x;
  }
  else {
    y= log(1.0+exp(x));
    if (!finite(y) | isnan(y))
      y=0.0;
  }
  return(y);
}

/* Softplus inverse function: y = log(exp(x)-1) */
double softplusinv(double x){
  double y;
  if(x>0.0){
    y = exp(x);
    if (isfinite(y) & !isnan(y))
      y=log(y-1.0);
    else
      y=x;
  }
  else {
    y= log(exp(x)-1.0);
  }
  return(y);
}


/* "Forward" pass in UMMH model */
void ummhfwd(double *theta, int *m, double *params){
  int i;
  double fprior=1.0;

  /* Apply transfer functions to mixture parameters. */
  for(i=*m-1;i>-1;i--){
    if(i>0){
      /* priors in [0,1] */
      params[i]=(1/(1+exp(-theta[4*i-1]))*(1.0-2*MINPI)+MINPI)*fprior;       
      fprior=fprior-params[i];
    }
    else
      params[i]=fprior;
    /* Rprintf("pi[%d] = %e \n",i+1,params[i]);*/
    params[*m+i]=softplus(theta[4*i])+MINXI;   /* tail parameters */
    params[2**m+i]=theta[4*i+1]; /* centers */
    params[3**m+i]=softplus(theta[4*i+2])+MINSIGMA; /* spread parameters */
  }

}

/* "Backward" pass in UMMH model */
void ummhbwd(double *params, int *m, double *theta){
  int i;
  double tmp,fprior=1.0;

  /* Apply inverse transfer functions to mixture parameters. */
  for(i=*m-1;i>-1;i--){
    if(i>0){
      tmp=params[i]/fprior;
      theta[4*i-1] = log(tmp/(1-tmp));
      fprior=fprior-params[i];
    }
    theta[4*i]=softplusinv(params[*m+i]-MINXI);  /* tail parameters */
    theta[4*i+1]=params[2**m+i]; /* centers */
    theta[4*i+2]=softplusinv(params[3**m+i]-MINSIGMA); /* spread parameters */
  }

}


/* Print the hybrid Pareto mixture model parameters */
void ummhprint(double *params, int *m){
  int j;
  /* double *params;
  params = (double*) R_alloc(4**m, sizeof(double));
  
  ummhfwd(theta,m,params);*/

  for (j=0; j<*m; j++){
    Rprintf("prior[%d] = %f, xi[%d] = %f, mu[%d] = %f, sigma[%d] = %f\n",
	      j+1,params[j],j+1,params[*m+j],j+1,params[2**m+j],j+1,params[3**m+j]);
  }

}


/* Hybrid Pareto mixture probability density function */
double ummhpdf(double *params, int m, double x){
  int j;
  double act,prob=0.0;;

  for(j=0; j<m; j++){     
    act=hpdf(params[m+j],params[2*m+j],params[3*m+j],x);
    prob+=params[j]*act;
  }		

  return(prob);
}


/* Hybrid Pareto mixture pdf at n points for use in R*/
void ummhpdfR(double *params, int *m, double *x, int *n, double *p){
  int i;

  for(i=0;i<*n;i++)
    p[i]=ummhpdf(params,*m,x[i]);
}


/* Hybrid Pareto mixture cumulative distribution function at n points*/
double ummhcdf(double *params, int m, double x){
  int j;
  double act,prob=0.0;

  for(j=0; j<m; j++){     
    act=hcdf(params[m+j],params[2*m+j],params[3*m+j],x);
    prob+=params[j]*act;
  }		
  return(prob);
}


/* Hybrid Pareto mixture cdf at n points for use in R*/
void ummhcdfR(double *params, int *m, double *x, int *n, double *p){
  int i;

  for(i=0;i<*n;i++)
    p[i]=ummhcdf(params,*m,x[i]);
}

/* Derivative of the hybrid Pareto pdf with respect to xi, mu and sigma 
   divided by the pdf: dhdxi/h */
void hpdfgrad(double xi, double mu, double sigma, double x,
	      double *dhdtheta){
  double w, gamma, alpha, beta, dh2dgdxi, dh2dgdbeta, dh2dgdalpha;
  
  w = lambertw((1.0 + xi)*(1.0 + xi)/(2.0*PI));
  gamma = 3.0/2.0 + erf(sign(1.0+xi)*sqrt(w/2.0))/2.0;
  alpha = sign(1.0+xi) * sigma * sqrt(w) + mu;
  
  /* Derivative with respect to xi, mu and sigma through gamma: dhdgammadxi. */
  dhdtheta[0] = -sqrt(w)*sign(1.0+xi)/
    (gamma*sqrt(2.0*PI)*exp(w/2.0)*(1.0+w)*(1.0+xi));	     /* dhdxi */
    
  if (x <= alpha){
    /* Derivative only depends on the Normal pdf */
    dhdtheta[1] = (x-mu)/ (sigma*sigma);			    /* dhdmu */
    dhdtheta[2] = (x-mu)*(x-mu)/(sigma*sigma*sigma)-1.0/sigma; /* dhdsigma */
  }
  else {
    /* Derivative only depends on the GP pdf */
    beta = sigma*fabs(1.0+xi)/sqrt(w);
    /* Compute the derivative of the GP pdf w/r to xi, beta and alpha. */
    if (xi != 0.0){
      dh2dgdxi=(1.0+xi)*(alpha-x)/(xi*(beta+xi*(x-alpha))) +
	log(1.0+ xi*(x-alpha)/beta)/(xi*xi);
      dh2dgdbeta=(x-beta-alpha)/(beta*(beta+ xi*(x-alpha)));
      dh2dgdalpha=(1.0+ xi)/(beta + xi*(x-alpha));
    }
    else{
      dh2dgdxi=0.0;
      dh2dgdbeta=(x-beta-alpha)/(beta*beta);
      dh2dgdalpha=1.0/beta;
    }
    /* Derivative w/r to xi has dependencies through GP
       pdf (xi, beta, alpha) */
    dhdtheta[0]+=(dh2dgdxi + dh2dgdbeta*(sign(1.0+ xi)*sigma/sqrt(w)-
					 sigma*fabs(1.0+xi)/(sqrt(w)*(1.0+w)*(1.0+xi)))+
		  dh2dgdalpha*sigma*sqrt(w)/((1.0+w)*fabs(1.0+xi)));  /* dhdxi */
    dhdtheta[1]=dh2dgdalpha;			      		  /* dhdmu */
    dhdtheta[2]=(dh2dgdalpha*sign(1.0+ xi)*sqrt(w) +
		  dh2dgdbeta*fabs(1.0+ xi)/sqrt(w));	       /* dhdsigma */
  }
}



/* Hybrid Pareto mixture negative log-likelihood at n points*/
void ummhnll(double *theta, int *m, double *x, int *n, double *nll, 
	       double *nllgrad){
  int i,j,nparams=4**m-1;
  double *params, *act, *post, *dhdtheta;
  double lprob,ppost,pprior,lprior; 
  
  
  /* Allocate memory for pointers. */
  params = (double*) R_alloc(4**m, sizeof(double));
  act = (double*) R_alloc(*m, sizeof(double));
  post = (double*) R_alloc(*m, sizeof(double));
  dhdtheta = (double*) R_alloc(3, sizeof(double));
  
	
  /* Make sure initialization of neg-log-like and its gradient
     is zero */
  *nll=0.0;
  for(j=0;j<nparams;j++)
    nllgrad[j]=0;


  /* Compute mixture parameters from vector theta */  
  ummhfwd(theta,m,params);

  for (i=0; i<*n; i++){
    /* Compute model log-probability for pattern i */
    /* mth component */
    act[*m-1] = hlogpdf(params[2**m-1], params[3**m-1], params[4**m-1],x[i]);

    if(theta[4**m-5]>0){
      pprior=-log(1+exp(-theta[4**m-5])); /* re-use computation */
      lprob=pprior+act[*m-1]; /* accumulates log(sum priors act) */
      post[*m-1]=lprob; /* log(prior) + log(act) */
      lprior=-theta[4**m-5]+pprior; /* accumulates sum log(1-g(theta_{4*j-1}) */
    }
    else{
      pprior=-log(1+exp(theta[4**m-5]));
      lprob=theta[4**m-5]+pprior+act[*m-1];
      post[*m-1]=lprob;
      lprior=pprior;
    }
    for(j=*m-2; j>-1; j--){
      act[j] = hlogpdf(params[*m+j], params[2**m+j], params[3**m+j],x[i]);

      if(j>0){
	if(theta[4*j-1]>0){
	  pprior=-log(1+exp(-theta[4*j-1]));
	  post[j]=pprior+lprior+act[j];
	  lprior+=-theta[4*j-1]+pprior;
	}
	else{
	  pprior=-log(1+exp(theta[4*j-1]));
	  post[j]=theta[4*j-1]+pprior+lprior+act[j];
	  lprior+=pprior;
	}
      }
      else 
	post[j]=lprior+act[j];
      
      if (lprob > post[j])
	lprob=lprob + log(1+exp(post[j]-lprob));
      else
	lprob=post[j] + log(1+exp(lprob-post[j]));            
    } /* for j */
 
    *nll -= lprob;
    
    /* Compute gradient for pattern i */
    ppost=0.0; /* Maintains \sum_{i=0}^{j-1} post[i] */
    pprior=0.0; /* Maintains \sum_{i=0}^{j-1} params[i] */
    for (j=0; j<*m; j++){
      /* Compute deltas for priors */
      post[j]=exp(post[j]-lprob); /* posterior */
      
      if(j > 0){ /* there are m-1 priors */
	ppost+=post[j-1];	
	nllgrad[4*j-1] += (params[j]*ppost-pprior*post[j])/
	  (pprior+params[j])*(1.0-2*MINPI);
      }
      pprior+=params[j];

      /* Gradient w/r to the hybrid Pareto parameters */
      hpdfgrad(params[*m+j], params[2**m+j],params[3**m+j],
	       x[i],dhdtheta);

      /* Compute gradients for hybrid Pareto parameters */
      nllgrad[4*j] += - post[j] *dhdtheta[0]*
	(1-exp(MINXI-params[*m+j]));   /* xi_j */
      /* deltas[4*j] = - post[j] *dhdtheta[0]*params[net->m+j]*
	 (1-params[net->m+j]);	 */
      nllgrad[4*j+1] += - post[j] * dhdtheta[1];    /* mu_j */
      nllgrad[4*j+2] += - post[j] * dhdtheta[2]
	*(1-exp(MINSIGMA-params[3**m+j]));	   /* sigma_j */
    } /* for j loop */
  } /* for i loop */
  
}



/* Hybrid Pareto mixture negative log-likelihood at n points with a penalty on the tail indexes based on a mixture (an exponential and a gaussian) with two modes: at zero and at 0.5 */
void ummhnll_bimodal_tailpen(double *theta, int *m, double * lambda, double *w,
			     double * beta, double *mu, double *sigma, double *x, int *n, 
			     double *nll,double *nllgrad){
			      			     
  int i,j,nparams=4**m-1;
  double *params, *act, *post, *dhdtheta;
  double lprob,ppost,pprior,lprior,tailpen,loga,logb; 
  
  
  /* Allocate memory for pointers. */
  params = (double*) R_alloc(4**m, sizeof(double));
  act = (double*) R_alloc(*m, sizeof(double));
  post = (double*) R_alloc(*m, sizeof(double));
  dhdtheta = (double*) R_alloc(3, sizeof(double));
	
  /* Make sure initialization of neg-log-like and its gradient
     is zero */
  *nll=0.0;
  for(j=0;j<nparams;j++)
    nllgrad[j]=0;


  /* Compute mixture parameters from vector theta */  
  ummhfwd(theta,m,params);

  for (i=0; i<*n; i++){
    /* Compute model log-probability for pattern i */
    /* mth component */
    act[*m-1] = hlogpdf(params[2**m-1], params[3**m-1], params[4**m-1],x[i]);

    if(theta[4**m-5]>0){
      pprior=-log(1+exp(-theta[4**m-5])); /* re-use computation */
      lprob=pprior+act[*m-1]; /* accumulates log(sum priors act) */
      post[*m-1]=lprob; /* log(prior) + log(act) */
      lprior=-theta[4**m-5]+pprior; /* accumulates sum log(1-g(theta_{4*j-1}) */
    }
    else{
      pprior=-log(1+exp(theta[4**m-5]));
      lprob=theta[4**m-5]+pprior+act[*m-1];
      post[*m-1]=lprob;
      lprior=pprior;
    }
    for(j=*m-2; j>-1; j--){
      act[j] = hlogpdf(params[*m+j], params[2**m+j], params[3**m+j],x[i]);

      if(j>0){
	if(theta[4*j-1]>0){
	  pprior=-log(1+exp(-theta[4*j-1]));
	  post[j]=pprior+lprior+act[j];
	  lprior+=-theta[4*j-1]+pprior;
	}
	else{
	  pprior=-log(1+exp(theta[4*j-1]));
	  post[j]=theta[4*j-1]+pprior+lprior+act[j];
	  lprior+=pprior;
	}
      }
      else 
	post[j]=lprior+act[j];
      
      if (lprob > post[j])
	lprob=lprob + log(1+exp(post[j]-lprob));
      else
	lprob=post[j] + log(1+exp(lprob-post[j]));            
    } /* for j */
 
    *nll -= lprob;
    
    /* Compute gradient for pattern i */
    ppost=0.0; /* Maintains \sum_{i=0}^{j-1} post[i] */
    pprior=0.0; /* Maintains \sum_{i=0}^{j-1} params[i] */
    for (j=0; j<*m; j++){
      /* Compute deltas for priors */
      post[j]=exp(post[j]-lprob); /* posterior */

      /* Rprintf("priors[%d] = %e \n",j+1,params[j]);*/
      
      if(j > 0){ /* there are m-1 priors */
	ppost+=post[j-1];	
	nllgrad[4*j-1] += (params[j]*ppost-pprior*post[j])/
	  (pprior+params[j])*(1.0-2*MINPI);
      }
      pprior+=params[j];

      /* Gradient w/r to the hybrid Pareto parameters */
      hpdfgrad(params[*m+j], params[2**m+j],params[3**m+j],
	       x[i],dhdtheta);

      /* Compute gradients for hybrid Pareto parameters */     
      nllgrad[4*j] += - post[j] *dhdtheta[0]*
	(1-exp(MINXI-params[*m+j]));   /* xi_j */
      /* deltas[4*j] = - post[j] *dhdtheta[0]*params[net->m+j]*
	 (1-params[net->m+j]);	 */
      nllgrad[4*j+1] += - post[j] * dhdtheta[1];    /* mu_j */
      nllgrad[4*j+2] += - post[j] * dhdtheta[2]
	*(1-exp(MINSIGMA-params[3**m+j]));	   /* sigma_j */
    } /* for j loop */
  } /* for i loop */

  /* Compute the tail penalty */
  for (j=0;j<*m;j++){
    loga=log(*w) +log(*beta) - *beta*params[*m+j];
    logb=log(1-*w)-(params[*m+j]-*mu)*(params[*m+j]-*mu)/(2**sigma**sigma) - log(2*PI)/2.0 - log(*sigma);
    *nll -= *lambda*(loga + log(1 + exp(logb-loga)));
    tailpen= *beta + ((params[*m+j]-*mu)/(*sigma**sigma)-*beta)/(exp(loga-logb)+1);
    nllgrad[4*j] += *lambda*tailpen*(1-exp(MINXI-params[*m+j])); 
  }


}




/* --------------------- UMM : Quantile computations  ----------------------*/

/* Bracket the quantile of level q in interval [a,b] given an initial 
   interval and the cdf of UMM. */
int ummqbrack(double (*cdf)(double *,int, double),
	      double *params, int m, double q, double *a, double *b){
  int j;
  double fa,fb;
  
  fa= (*cdf)(params,m,*a)-q;
  fb=(*cdf)(params,m,*b)-q;

  /*Rprintf("fa = %f\n",fa);
    Rprintf("fb = %f\n",fb);*/

  for(j=0;j<NTRY;j++){
    if(fa*fb < 0.0) return 1;
    if(fabs(fa) < fabs(fb)){
      *a+=FACTOR*(*a-*b);
      fa= (*cdf)(params,m,*a)-q;
    }
    else{
      *b+=FACTOR*(*b-*a);
      fb=(*cdf)(params,m,*b)-q;
    }
  }
  return 0;
}


/* Quantile computation given a cdf and pdf of UMM:
   numerical approximation by bisection and Newton's methods,
   first bracketing interval given by [a,b] */
void ummquant(double (*cdf)(double *,int, double),
	       double (*pdf)(double *,int, double),
	       double *params, int m, double q, double *aa, double *bb,
	       double tol, int itmax, double *xq){ 
  
  int j, brack;
  double xqi, fa,fq,a,b;


  /* Bracket the quantile. */
  brack=ummqbrack(cdf,params, m, q, aa, bb);
  a=*aa;
  b=*bb;
  /*mexPrintf("a = %f\n",a); */
  /* mexPrintf("b = %f\n",b); */

   if (brack == 1){ /* bracketing succeeded */
    /* start a few iterations of the bisection method */
    j=0;
    fa= (*cdf)(params,m,a)-q;
    while(j<IBIS){ /* IBIS = 100*/
      *xq=a+(b-a)/2;
      fq=(*cdf)(params,m,*xq)-q;
      if((fq==0.0) | ((b-a)/2 < tol))
	break;
      j++;
      if(fa*fq>0){
	a=*xq;
	fa=fq;
      }
      else
	b=*xq;
      /* mexPrintf("xq = %f\n",*xq);*/
    } /* bisection loop */

    j=0;
    while(j<itmax){ /* Newton's method with initial value *xq */
      xqi=*xq-((*cdf)(params,m,*xq)-q)/(*pdf)(params,m,*xq);
      if(fabs(xqi-*xq)/fabs(xqi)<tol){
	*xq=xqi;
	return;
      }
      j++;
      *xq=xqi;
    } /* Newton loop */
   } /* successful bracket if */
   else{
     Rprintf("Bracketing failed\n");
     *xq=NAN;
   }
}

/* Gaussian mixture probability density function at n points*/
double ummgpdf(double *params, int m, double x){
  int j;
  double act,prob=0.0;

  for(j=0; j<m; j++){     
    act=normpdf(params[m+j],params[2*m+j],x);
    prob+=params[j]*act;
  }		

  return(prob);
}

/* Gaussian mixture pdf at n points for use in R*/
void ummgpdfR(double *params, int *m, double *x, int *n, double *p){
  int i;

  for(i=0;i<*n;i++)
    p[i]=ummgpdf(params,*m,x[i]);
}


/* Gaussian mixture cumulative distribution function */
double ummgcdf(double *params, int m, double x){
  int j;
  double act,prob=0.0;

  for(j=0; j<m; j++){     
    act=normcdf(params[m+j],params[2*m+j],x);
    prob+=params[j]*act;
  }		
  return(prob);
}

/* Gaussian mixture cdf at n points for use in R*/
void ummgcdfR(double *params, int *m, double *x, int *n, double *p){
  int i;

  for(i=0;i<*n;i++)
    p[i]=ummgcdf(params,*m,x[i]);
}



/* "Backward" pass in UMMG model */
void ummgbwd(double *params, int *m, double *theta){
  int i;
  double tmp,fprior=1.0;

  /* Apply inverse transfer functions to mixture parameters. */
  for(i=*m-1;i>-1;i--){
    if(i>0){
      tmp=params[i]/fprior;
      theta[3*i-1] = log(tmp/(1-tmp));
      fprior=fprior-params[i];
    }
    theta[3*i]=params[*m+i]; /* centers */
    theta[3*i+1]=softplusinv(params[2**m+i]-MINSIGMA); /* spread parameters */
  }

}
